/*
  Lessfs: A data deduplexing filesystem.
  Copyright (C) 2008 Mark Ruijter <mruijter@lessfs.com>
*/

#define FUSE_USE_VERSION 26
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <malloc.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <signal.h>
#include <libgen.h>
#include <sys/utsname.h>
#include <sys/vfs.h>

#include <tcutil.h>
#include <tchdb.h>
#include <tcbdb.h>
#include <stdlib.h>
#include <stdbool.h>
#ifdef LZO
#include "lib_lzo.h"
#else
#include "lib_qlz.h"
#endif
#include "lib_tc.h"
#include "lib_net.h"

#ifdef HAVE_SETXATTR
#include <sys/xattr.h>
#endif
#include "lib_log.h"
#include "lib_safe.h"
#include "lib_cfg.h"
#include "lib_str.h"
#include "retcodes.h"

#ifdef i386
#define ITERATIONS 30
#else
#define ITERATIONS 500
#endif

typedef unsigned long long int word64;
typedef unsigned long word32;
typedef unsigned char byte;
void tiger(byte *, word64, word64 *);

struct configdata *config;

char *logname;
char *function;
int debug = 5;
int BLKSIZE = 4096;
#define MAX_THREADS 1
int max_threads = MAX_THREADS;
BLKDTA **tdta = NULL;

extern TCHDB *dbp;
extern TCHDB *dbu;
extern TCHDB *dbb;
extern TCHDB *dbdta;
extern TCHDB *dbs;
extern TCBDB *dbdirent;
extern TCBDB *dbl;
extern TCMDB *dbcache;
extern TCMDB *blkcache;

BLKDTA *blkdta;

void segvExit()
{
    LFATAL("Exit caused by segfault!, exitting\n");
    tc_close(0);
    exit(EXIT_OK);
}

void normalExit()
{
    LFATAL("Exit signal received, exitting\n");
    tc_close(0);
    exit(EXIT_OK);
}

void libSafeExit()
{
    tc_close(0);
    LFATAL("Exit signal received from lib_safe, exitting\n");
    exit(EXIT_OK);
}

int check_path_sanity(const char *path)
{
    char *name;
    char *p;
    char *q;
    int retcode = 0;

    FUNC;
    LDEBUG("check_path_sanity : %s", path);

    name = s_strdup((char *) path);
    p = name;
    q = p;
    while (1) {
        p = strrchr(q, '/');
        if (p == NULL)
            break;
        q = p;
        q++;
        p[0] = 0;
        if (q) {
            if (strlen(q) > 255) {
                LDEBUG("check_path_sanity : error q>255");
                retcode = -ENAMETOOLONG;
            }
        }
        q = p--;
    }
    free(name);
    EFUNC;
    return (retcode);
}

void dbsync()
{
    tcbdbsync(dbdirent);
    tchdbsync(dbu);
    tchdbsync(dbb);
    tchdbsync(dbp);
    tchdbsync(dbdta);
}

static int lessfs_getattr(const char *path, struct stat *stbuf)
{
    int res;

    FUNC;
    LDEBUG("lessfs_getattr %s", path);
    res = check_path_sanity(path);
    if (res == -ENAMETOOLONG)
        return (res);

    getLock();
    res = dbstat(path, stbuf);
    LDEBUG("lessfs_getattr : st_nlinks=%i", stbuf->st_nlink);
    LDEBUG("-Got size %llu : result %i",
           (unsigned long long) stbuf->st_size, res);
    releaseLock();
    return (res);
}

static int lessfs_access(const char *path, int mask)
{
    int res = 0;
    FUNC;
// Empty stub
    return (res);
}

static int lessfs_readlink(const char *path, char *buf, size_t size)
{
    int res = 0;

    FUNC;

    LDEBUG("lessfs_readlink : size = %i", (int) size);
    getLock();
    res = fs_readlink(path, buf, size);
    releaseLock();
    return (res);
}


static int lessfs_readdir(const char *path, void *buf,
                          fuse_fill_dir_t filler, off_t offset,
                          struct fuse_file_info *fi)
{
    int retcode;
    getLock();
    retcode = fs_readdir(path, buf, filler, offset, fi);
    releaseLock();
    return (retcode);
}

static int lessfs_mknod(const char *path, mode_t mode, dev_t rdev)
{
    int retcode = 0;
    char *dname;
    struct stat stbuf;
    time_t thetime;

    FUNC;
    getLock();
    thetime = time(NULL);
    dname = s_dirname((char *) path);
    retcode = dbstat(dname, &stbuf);
    LDEBUG("lessfs_mknod : dbstat returns %i", retcode);
    if (0 != retcode)
        return retcode;
    dbmknod(path, mode, NULL, rdev);
    stbuf.st_ctime = thetime;
    stbuf.st_mtime = thetime;
    retcode = update_stat(dname, &stbuf);
    LDEBUG("lessfs_mknod : update_stat returns %i", retcode);
    free(dname);
    releaseLock();
    return (retcode);
}

static int lessfs_mkdir(const char *path, mode_t mode)
{
    int ret;
    getLock();
    ret = fs_mkdir(path, mode);
    releaseLock();
    return (ret);
}

static int lessfs_unlink(const char *path)
{
    int res;
    getLock();
    res = unlink_file(path);
    dbsync();
    releaseLock();
    return res;
}

static int lessfs_rmdir(const char *path)
{
    int res;
    getLock();
    res = fs_rmdir(path);
    releaseLock();
    return (res);

}

static int lessfs_symlink(const char *from, const char *to)
{
    int res = 0;

    FUNC;
    getLock();
    res = fs_symlink((char *) from, (char *) to);
    releaseLock();
    return (res);
}

static int lessfs_rename(const char *from, const char *to)
{
    int res = 0;
    struct stat stbuf;

    FUNC;
    LDEBUG("lessfs_rename : from %s , to %s", from, to);
    getLock();
    res = dbstat(from, &stbuf);
    if (res == 0) {
        update_filesize_onclose(stbuf.st_ino);
        if (stbuf.st_nlink > 1) {
            res = fs_rename_link(from, to, stbuf);
        } else {
            res = fs_rename(from, to, stbuf);
        }
    }
    releaseLock();
    return (res);
}

static int lessfs_link(const char *from, const char *to)
{
    int res = 0;

    FUNC;
    getLock();
    res = fs_link((char *) from, (char *) to);
    releaseLock();
    return (res);
}

static int lessfs_chmod(const char *path, mode_t mode)
{
    int res;
    time_t thetime;
    struct stat stbuf;

    FUNC;
    LDEBUG("lessfs_chmod : %s", path);
    getLock();
    thetime = time(NULL);
    res = dbstat(path, &stbuf);
    if (res == 0) {
        stbuf.st_mode = mode;
        stbuf.st_ctime = thetime;
        res = update_stat((char *) path, &stbuf);
    }
    releaseLock();
    return (res);
}


static int lessfs_chown(const char *path, uid_t uid, gid_t gid)
{
    int res;
    time_t thetime;
    struct stat stbuf;

    FUNC;

    getLock();
    thetime = time(NULL);
    res = dbstat(path, &stbuf);
    if (res == 0) {
        if (-1 != uid)
            stbuf.st_uid = uid;
        if (-1 != gid)
            stbuf.st_gid = gid;
        stbuf.st_ctime = thetime;
        res = update_stat((char *) path, &stbuf);
    }
    releaseLock();
    return (res);
}

static int lessfs_truncate(const char *path, off_t size)
{
    int res = 0;
    struct stat *stbuf;
    char *bname;
    
    getLock();
    bname = s_basename((char *) path);
    stbuf=s_malloc(sizeof(struct stat));
    res = dbstat(path, stbuf);
    if (res != 0)
        return (res);
    if (S_ISDIR(stbuf->st_mode)) {
        return (-EISDIR);
    }
    res = fs_truncate(stbuf, size, bname);
    free(stbuf);
    free(bname);
    releaseLock();
    return (res);
}

static int lessfs_utimens(const char *path, const struct timespec ts[2])
{
    int res = 0;
    struct timeval tv[2];
    struct stat stbuf;
    DDSTAT *ddstat;
    DBT *ddbuf;
    DBT *dataptr;

    FUNC;
    tv[0].tv_sec = ts[0].tv_sec;
    tv[1].tv_sec = ts[1].tv_sec;

    getLock();
    res = dbstat(path, &stbuf);
    if (res == 0) {
        dataptr =
            search_dbdata(dbp, &stbuf.st_ino, sizeof(unsigned long long));
        if (dataptr == NULL) {
#ifdef x86_64
            die_dataerr("Failed to find file %lu", stbuf.st_ino);
#else
            die_dataerr("Failed to find file %llu", stbuf.st_ino);
#endif
        }
        ddstat = value_to_ddstat((char *) dataptr->data, dataptr->size);
        ddstat->stbuf.st_atime = tv[0].tv_sec;
        ddstat->stbuf.st_mtime = tv[1].tv_sec;
        ddbuf = create_ddbuf(ddstat->stbuf, ddstat->filename);
        bin_write_dbdata(dbp, &stbuf.st_ino, sizeof(unsigned long long),
                         (void *) ddbuf->data, ddbuf->size);
        DBTfree(dataptr);
        DBTfree(ddbuf);
        ddstatfree(ddstat);
    }
    releaseLock();
    return (res);
}

static int lessfs_open(const char *path, struct fuse_file_info *fi)
{
    struct stat stbuf;
    char *bname;
    DBT *ddbuf;
    DBT *dataptr;
    unsigned long long blocknr = 0;
    MEMDDSTAT *ddstat;
    unsigned long long inode;

    int res = 0;

    FUNC;
    LDEBUG("lessfs_open : %s strlen %i : uid %u", path,
           strlen((char *) path), fuse_get_context()->uid);

    open_lock();
    getLock();
    res = dbstat(path, &stbuf);
    releaseLock();
    if (res == -ENOENT) {
        fi->fh = 0;
        stbuf.st_mode = fi->flags;
        stbuf.st_uid = fuse_get_context()->uid;
        stbuf.st_gid = fuse_get_context()->gid;
    } else {
        fi->fh = stbuf.st_ino;
        inode = stbuf.st_ino;
        wait_io_pending(inode);
        bname = s_basename((char *) path);
//  Check if we have already something in cache for this inode
        dataptr =
            search_memhash(dbcache, &stbuf.st_ino,
                           sizeof(unsigned long long));
        if (dataptr == NULL) {
            blocknr--;          /* Invalid block in cache */
//  Initialize the cache
            ddstat = s_malloc(sizeof(MEMDDSTAT));
            memcpy(&ddstat->stbuf, &stbuf, sizeof(struct stat));
            memcpy(&ddstat->filename, bname, strlen(bname) + 1);
            ddstat->blocknr = blocknr;
            ddstat->updated = 0;
            ddstat->opened = 1;
            ddstat->deduplicated = 0;
            ddstat->lzo_compressed_size = 0;
            LDEBUG("lessfs_open : initial open ddstat->opened = %u",
                   ddstat->opened);
            ddbuf = create_mem_ddbuf(ddstat);
            mbin_write_dbdata(dbcache, &inode, sizeof(unsigned long long),
                              (void *) ddbuf->data, ddbuf->size);
            DBTfree(ddbuf);
        } else {
            ddstat =
                value_tomem_ddstat((char *) dataptr->data, dataptr->size);
            ddstat->opened++;
            LDEBUG("lessfs_open : repeated open ddstat->opened = %u",
                   ddstat->opened);
            ddbuf = create_mem_ddbuf(ddstat);
            mbin_write_dbdata(dbcache, &inode, sizeof(unsigned long long),
                              (void *) ddbuf->data, ddbuf->size);
            DBTfree(dataptr);
        }
        memddstatfree(ddstat);
        free(bname);
        releaseLock();
    }
    release_open_lock();
    return (res);
}

static int lessfs_read(const char *path, char *buf, size_t size,
                       off_t offset, struct fuse_file_info *fi)
{
    unsigned long long blocknr;
    size_t done = 0;
    size_t got = 0;
    size_t block_offset = 0;
    char *tmpbuf = NULL;

    FUNC;
    getLock();
    LDEBUG("lessfs_read called offset : %llu, size bytes %llu",
           (unsigned long long) offset, (unsigned long long) size);
    tmpbuf = s_malloc(BLKSIZE + 1);
    memset(tmpbuf, 0, BLKSIZE + 1);
    while (done < size) {
        blocknr = (offset + done) / BLKSIZE;
        LDEBUG("blocknr to read :%llu", (unsigned long long) blocknr);
        block_offset = (done + offset) - (blocknr * BLKSIZE);
        got = readBlock(blocknr, path, tmpbuf, fi->fh);
        if (got > 0) {
            if ((got - block_offset) + done >= size) {
                memcpy(buf + done, tmpbuf + block_offset, size - done);
                done = size;
                break;
            }
            memcpy(buf + done, tmpbuf + block_offset, got - block_offset);
            done = done + got - block_offset;
        } else
            break;
    }
    if (done == 0)
        memset(buf, 0, size);
    free(tmpbuf);
    releaseLock();
    return (size);
}

// Return 0 to add the block or 1 to update the block
int update_or_addblock(unsigned long long inode,
                       unsigned long long blocknr)
{
    int res = 0;
    DBT *dataptr;
    MEMDDSTAT *ddstat;
    unsigned long long fsize;

    fsize = blocknr * BLKSIZE;
    dataptr = search_memhash(dbcache, &inode, sizeof(unsigned long long));
    if (dataptr == NULL)
        return (res);
    ddstat = value_tomem_ddstat((char *) dataptr->data, dataptr->size);
    if (fsize < ddstat->stbuf.st_size)
        res = 1;
    DBTfree(dataptr);
    memddstatfree(ddstat);
    return (res);
}


static int lessfs_write(const char *path, const char *buf, size_t size,
                        off_t offset, struct fuse_file_info *fi)
{
    unsigned long long blocknr;
    unsigned int offsetblock;
    DBT *blocktiger;
    size_t bsize;
    size_t done = 0;
    int res;
    INOBNO inobno;
    int pending;

    tiger_lock();
    bsize = size;
    blocknr = offset / BLKSIZE;
    offsetblock = offset - (blocknr * BLKSIZE);
    if ((offsetblock + bsize) > BLKSIZE) {
        bsize = BLKSIZE - offsetblock;
    }
    blkdta->inode = fi->fh;
    inobno.inode = fi->fh;
  wagain:
    blkdta->blockdata = s_malloc(bsize);
    inobno.blocknr = blocknr;
/*  When I/O for this inode - blocknr is pending this operation will be an update */
    getLock();
    pending = inode_block_pending(blkdta->inode, blocknr);
    if (0 == pending) {
        res = update_or_addblock(blkdta->inode, blocknr);
    } else {
        LDEBUG
            ("lessfs_write : Pending I/O for inode - blocknr %llu - %llu, update_block called",
             blkdta->inode, blocknr);
        res = 1;
    }
    releaseLock();
    if (1 == res) {
        wait_io_pending(blkdta->inode);
        blocktiger = check_block_exists(inobno);
        releaseLock();
        if (NULL != blocktiger) {
            getLock();
            LDEBUG("lessfs_write -> update_block : inode %llu blocknr %llu", fi->fh, blocknr);
            update_block(buf + done, blocknr, offsetblock,
                         bsize, blkdta->inode, blocktiger->data);
            DBTfree(blocktiger);
            releaseLock();
        } else {
            LDEBUG("lessfs_write -> add_block : inode %llu blocknr %llu", fi->fh, blocknr);
            blkdta->blocknr = blocknr;
            blkdta->offsetblock = offsetblock;
            blkdta->bsize = bsize;
            blkdta->sparse = 1;
            memcpy((char *) blkdta->blockdata, buf + done, bsize);
            thread_lock_release();
            write_lock();
        }
    } else {
        LDEBUG("lessfs_write -> add_block : inode %llu blocknr %llu", fi->fh, blocknr);
        blkdta->blocknr = blocknr;
        blkdta->offsetblock = offsetblock;
        blkdta->bsize = bsize;
        blkdta->sparse = 0;
        memcpy((char *) blkdta->blockdata, buf + done, bsize);
        thread_lock_release();
        write_lock();
    }
    done = done + bsize;
    bsize = size - done;
    free((char *) blkdta->blockdata);
    if (done < size) {
        blocknr++;
        offsetblock = 0;
        goto wagain;
    }
    release_tiger_lock();
    return (done);
}

static int lessfs_statfs(const char *path, struct statvfs *stbuf)
{
    int res;

    res = statvfs(config->blockdata, stbuf);
    if (res == -1)
        return -errno;
    return 0;
}

static int lessfs_release(const char *path, struct fuse_file_info *fi)
{
    DBT *dataptr;
    MEMDDSTAT *ddstat;
    DBT *ddbuf;
    DBT *data;

    FUNC;
// Finish pending i/o for this inode.
    wait_io_pending(fi->fh);
// Flush blocks in cache that where not written yet, if any.
    data = try_block_cache(fi->fh, 0, 1);
    if (data != NULL)
        die_syserr();
// Update the filesize when needed.
    update_filesize_onclose(fi->fh);

    dataptr = search_memhash(dbcache, &fi->fh, sizeof(unsigned long long));
    if (dataptr != NULL) {
        ddstat = value_tomem_ddstat((char *) dataptr->data, dataptr->size);
#ifdef x86_64
        LFATAL
            ("File %s size %lu bytes : lzo compressed bytes %llu and %llu duplicate blocks",
             path, ddstat->stbuf.st_size, ddstat->lzo_compressed_size,
             ddstat->deduplicated);
#else
        LFATAL
            ("File %s size %llu bytes : lzo compressed bytes %llu and %llu duplicate blocks",
             path, ddstat->stbuf.st_size, ddstat->lzo_compressed_size,
             ddstat->deduplicated);
#endif
        LFATAL("File %s compression ratio : 1 : %f", path,
               (float) ddstat->stbuf.st_size /
               ddstat->lzo_compressed_size);
        if (ddstat->opened == 1) {
            LDEBUG("lessfs_release : Delete cache for %llu", fi->fh);
            mdelete_key(dbcache, &fi->fh, sizeof(unsigned long long));
        } else {
            ddstat->opened--;
            ddbuf = create_mem_ddbuf(ddstat);
            LDEBUG("lessfs_release : %llu is now %u open", fi->fh,
                   ddstat->opened);
            mbin_write_dbdata(dbcache, &fi->fh, sizeof(unsigned long long),
                              (void *) ddbuf->data, ddbuf->size);
            DBTfree(ddbuf);
        }
        memddstatfree(ddstat);
        DBTfree(dataptr);
    }
    dbsync();
    (void) path;
    (void) fi;
    releaseLock();
    return 0;
}

static int lessfs_fsync(const char *path, int isdatasync,
                        struct fuse_file_info *fi)
{
    DBT *data;
    FUNC;
    (void) path;
    (void) isdatasync;
    wait_io_pending(fi->fh);
    data = try_block_cache(fi->fh, 0, 1);
    if (data != NULL)
        die_syserr();
    update_filesize_onclose(fi->fh);
    dbsync();
    releaseLock();
    return 0;
}

static void lessfs_destroy(void *unused __attribute__ ((unused)))
{
    FUNC;
    tc_close(0);
    free(config);
    return;
}

void freeze_nospace(char *dbpath)
{
   struct statfs sb;
   LFATAL("Filesystem for database %s has insufficient space to continue, freezing I/O",dbpath);
   getLock();
   tc_close(1);
   LFATAL("All IO is now suspended until space comes available."); 
   LFATAL("If no other options are available it should be safe to kill lessfs.");
   while(1) {
      if ( -1 == statfs(dbpath,&sb)) die_dataerr("Failed to stat : %s",dbpath);
      if ( 0 == check_free_space(&sb)) {
         tc_open(1);
         releaseLock();
         break;
      }
      sleep(1);
   }
}

int check_free_space(struct statfs *sb)
{
   float dfull; 
   int mf;
   char *minfree;

   dfull=(float)sb->f_blocks/(float)sb->f_bfree;
   dfull=100/dfull;
   minfree=getenv("MINSPACEFREE");
   if ( NULL != minfree ) {
     mf=atoi(minfree);
     if ( mf > 100 || mf < 1 ) mf=10;
   } else {
     mf=10;
   }
   if ( dfull <= mf ){
       LFATAL("check_free_space : %f only procent free space left",dfull);
       return(-1);
   } 
   return(0);
}

/* This thread does general housekeeping.
   For now it checks that there is enough diskspace free for
   the databases. If not it syncs the database, 
   closes them and freezes I/O. */
void *housekeeping_worker(void *arg)
{
    int interval;
    struct statfs sb;
    char *dbpath;
    int freeze=0;
    char *iv;
 
    iv=getenv("INSPECT_DISK_INTERVAL");
    if ( NULL == iv ) {
       interval=1;
    } else {
       interval=atoi(iv);
    } 
    
    while(1){
       dbpath = as_sprintf("%s/fileblock.tch", config->fileblock);
       if ( -1 == statfs(dbpath,&sb)) die_dataerr("Failed to stat : %s",dbpath);
       if ( 0 != check_free_space(&sb)) freeze_nospace(dbpath);
       free(dbpath);
       dbpath = as_sprintf("%s/blockusage.tch", config->blockusage);
       if ( -1 == statfs(dbpath,&sb)) die_dataerr("Failed to stat : %s",dbpath);
       if ( 0 != check_free_space(&sb)) freeze_nospace(dbpath);
       free(dbpath);
       dbpath = as_sprintf("%s/metadata.tcb", config->meta);
       if ( -1 == statfs(dbpath,&sb)) die_dataerr("Failed to stat : %s",dbpath);
       if ( 0 != check_free_space(&sb)) freeze_nospace(dbpath);
       free(dbpath);
       dbpath = as_sprintf("%s/blockdata.tch", config->blockdata); 
       if ( -1 == statfs(dbpath,&sb)) die_dataerr("Failed to stat : %s",dbpath);
       if ( 0 != check_free_space(&sb)) freeze_nospace(dbpath);
       free(dbpath);
       dbpath = as_sprintf("%s/symlink.tch", config->symlink);
       if ( -1 == statfs(dbpath,&sb)) die_dataerr("Failed to stat : %s",dbpath);
       if ( 0 != check_free_space(&sb)) freeze_nospace(dbpath);
       free(dbpath);
       dbpath = as_sprintf("%s/dirent.tcb", config->dirent);
       if ( -1 == statfs(dbpath,&sb)) die_dataerr("Failed to stat : %s",dbpath);
       if ( 0 != check_free_space(&sb)) freeze_nospace(dbpath);
       free(dbpath);
       dbpath = as_sprintf("%s/hardlink.tcb", config->hardlink);
       if ( -1 == statfs(dbpath,&sb)) die_dataerr("Failed to stat : %s",dbpath);
       if ( 0 != check_free_space(&sb)) freeze_nospace(dbpath);
       free(dbpath);
       sleep(interval);
    }
}

void *ioctl_worker(void *arg)
{
    int msocket;
    int csocket;
    const char *port;
    const char *proto = "tcp";
    struct sockaddr_un client_address;
    socklen_t client_len;
    char buf[1028];
    char *addr;
    int err = 0;
    char *result;
    char *message = NULL;
    bool isfrozen = 0;

    msocket = -1;
    while (1) {
        addr = getenv("LISTEN_IP");
        port = getenv("LISTEN_PORT");
        if (NULL == port)
            port = "100";
        if (NULL == addr)
            LWARNING
                ("The administration session is not limited to localhost, this is not recommended.");
        msocket = serverinit(addr, port, proto);
        if (msocket != -1)
            break;
        sleep(1);
        close(msocket);
    }

    client_len = sizeof(client_address);
    while (1) {
        csocket =
            accept(msocket, (struct sockaddr *) &client_address,
                   &client_len);
        while (1) {
            result = NULL;
            if (-1 == timeoutWrite(3, csocket, ">", 1))
                break;
            if (-1 == readnlstring(10, csocket, buf, 1024)) {
                result = "timeout";
                err = 1;
                break;
            }
            if (0 == strncmp(buf, "\r", strlen("\r")))
                continue;
            if (0 == strncmp(buf, "quit\r", strlen("quit\r"))
                || 0 == strncmp(buf, "exit\r", strlen("exit\r"))) {
                err = 0;
                result = "bye";
                break;
            }
            if (0 == strncmp(buf, "freeze\r", strlen("freeze\r"))) {
                if (0 == isfrozen) {
                    result = "All i/o is now suspended.";
                    err = 0;
                    getLock();
                    dbsync();
                    isfrozen = 1;
                } else {
                    result = "i/o is already suspended.";
                    err = 1;
                }
            }
            if (0 == strncmp(buf, "defrost\r", strlen("defrost\r"))) {
                if (1 == isfrozen) {
                    result = "Resuming i/o.";
                    err = 0;
                    releaseLock();
                    isfrozen = 0;
                } else {
                    result = "i/o is not suspended.";
                    err = 1;
                }
            }
            if (0 == strncmp(buf, "defrag\r", strlen("defrag\r"))) {
                result = "Resuming i/o after defragmentation.";
                err = 1;
                if (-1 ==
                    timeoutWrite(3, csocket,
                                 "Suspending i/o for defragmentation\n",
                                 strlen
                                 ("Suspending i/o for defragmentation\n")))
                    break;
                err = 0;
                getLock();
                tc_defrag();
                tc_close(1);
                tc_open(1);
                releaseLock();
            }
            if (0 == strncmp(buf, "help\r", strlen("help\r"))
                || 0 == strncmp(buf, "h\r", strlen("h\r"))) {
                result =
                    "valid commands: defrag defrost freeze help quit|exit";
                err = 0;
            }
            if (NULL == result) {
                err = -1;
                result = "unknown command";
            }
            if (err == 0) {
                message = as_sprintf("+OK %s\n", result);
                if (-1 ==
                    timeoutWrite(3, csocket, message, strlen(message)))
                    break;
            } else {
                message = as_sprintf("-ERR %s\n", result);
                if (-1 ==
                    timeoutWrite(3, csocket, message, strlen(message)))
                    break;
            }
            free(message);
            message = NULL;
        }
        if (message)
            free(message);
        if (err == 0) {
            message = as_sprintf("+OK %s\n", result);
            timeoutWrite(3, csocket, message, strlen(message));
        } else {
            message = as_sprintf("-ERR %s\n", result);
            timeoutWrite(3, csocket, message, strlen(message));
        }
        free(message);
        message = NULL;
        close(csocket);
    }
    return NULL;
}

void *init_worker(void *arg)
{
    int count;
    int c;

    memcpy(&count, arg, sizeof(int));
    getLock();
    if (NULL == tdta) {
        tdta = malloc(max_threads * sizeof(BLKDTA));
        for (c = 0; c < max_threads; c++) {
            tdta[c] = s_malloc(sizeof(BLKDTA));
            tdta[c]->blockfiller = s_malloc(BLKSIZE);
            tdta[c]->inode = 0;
        }
    }
    releaseLock();

    while (1) {
        thread_lock();
        dta_lock(count);
        tdta[count]->blockdata = s_malloc(blkdta->bsize);
        memcpy((char *) tdta[count]->blockdata, (char *) blkdta->blockdata,
               blkdta->bsize);
        memcpy(&tdta[count]->bsize, &blkdta->bsize, sizeof(size_t));
        getLock();
        tdta[count]->inode = blkdta->inode;
        memcpy(&tdta[count]->blocknr, &blkdta->blocknr,
               sizeof(unsigned long long));
        releaseLock();
        memcpy(&tdta[count]->sparse, &blkdta->sparse, sizeof(bool));
        memcpy(&tdta[count]->offsetblock, &blkdta->offsetblock,
               sizeof(off_t));
        release_write_lock();
        memset(tdta[count]->blockfiller, 0, BLKSIZE);
        memmove(tdta[count]->blockfiller + tdta[count]->offsetblock,
                tdta[count]->blockdata, tdta[count]->bsize);
        tdta[count]->stiger = binhash(tdta[count]->blockfiller, BLKSIZE);
        tdta[count]->compressed = NULL;
      iopending:
        for (c = 0; c < max_threads; c++) {
            if (c == count)
                continue;
            if (0 == tdta[c]->inode)
                continue;
            if (tdta[c]->inode == tdta[count]->inode) {
                if (tdta[c]->blocknr < tdta[count]->blocknr) {
                    if (NULL == tdta[count]->compressed) {
#ifdef LZO
                        tdta[count]->compressed =
                            lzo_compress(tdta[count]->blockfiller,
                                         BLKSIZE);
#else
                        tdta[count]->compressed =
                            clz_compress(tdta[count]->blockfiller,
                                         BLKSIZE);
#endif
                    }
                    goto iopending;
                }
            }
        }
        getLock();
        addBlock(tdta[count]);
        tdta[count]->inode = 0;
        releaseLock();
        free((char *) tdta[count]->blockdata);
        free(tdta[count]->stiger);
        release_dta_lock(count);
    }
    free(tdta[count]->blockfiller);
    free(tdta[count]);
    tdta[count] = NULL;
    pthread_exit(NULL);
}


static void *lessfs_init()
{
    unsigned int count = 0;
    unsigned int *cnt;
    int res;

    FUNC;
    config = s_malloc(sizeof(struct configdata));
    config->blockdata = read_val("BLOCKDATA_PATH");
    config->blockdatabs = read_val("BLOCKDATA_BS");
    config->blockusage = read_val("BLOCKUSAGE_PATH");
    config->blockusage = read_val("BLOCKUSAGE_PATH");
    config->blockusagebs = read_val("BLOCKUSAGE_BS");
    config->dirent = read_val("DIRENT_PATH");
    config->direntbs = read_val("DIRENT_BS");
    config->fileblock = read_val("FILEBLOCK_PATH");
    config->fileblockbs = read_val("FILEBLOCK_BS");
    config->meta = read_val("META_PATH");
    config->metabs = read_val("META_BS");
    config->hardlink = read_val("HARDLINK_PATH");
    config->hardlinkbs = read_val("HARDLINK_BS");
    config->symlink = read_val("SYMLINK_PATH");
    config->symlinkbs = read_val("SYMLINK_BS");
    if (NULL == dbp)
        tc_open(0);
#ifdef LZO
    initlzo();
#endif
    blkdta = s_malloc(sizeof(BLKDTA));
    if (NULL != getenv("MAX_THREADS"))
        max_threads = atoi(getenv("MAX_THREADS"));

    pthread_t worker_thread[max_threads];
    pthread_t ioctl_thread;
    pthread_t housekeeping_thread;

    LDEBUG("lessfs_init : thread_lock");
    thread_lock();
    LDEBUG("lessfs_init : write_lock");
    write_lock();
    for (count = 0; count < max_threads; count++) {
        cnt = s_malloc(sizeof(int));
        memcpy(cnt, &count, sizeof(int));
        dta_mutex_init(count);
        res =
            pthread_create(&(worker_thread[count]), NULL, init_worker,
                           (void *) cnt);
        if (res != 0)
            die_syserr();
    }
    res = pthread_create(&ioctl_thread, NULL, ioctl_worker, (void *) NULL);
    if (res != 0)
        die_syserr();
    res = pthread_create(&housekeeping_thread, NULL, housekeeping_worker, (void *) NULL);
    if (res != 0)
        die_syserr();
    return NULL;
}

static struct fuse_operations lessfs_oper = {
    .getattr = lessfs_getattr,
    .access = lessfs_access,
    .readlink = lessfs_readlink,
    .readdir = lessfs_readdir,
    .mknod = lessfs_mknod,
    .mkdir = lessfs_mkdir,
    .symlink = lessfs_symlink,
    .unlink = lessfs_unlink,
    .rmdir = lessfs_rmdir,
    .rename = lessfs_rename,
    .link = lessfs_link,
    .chmod = lessfs_chmod,
    .chown = lessfs_chown,
    .truncate = lessfs_truncate,
    .utimens = lessfs_utimens,
    .open = lessfs_open,
    .read = lessfs_read,
    .write = lessfs_write,
    .statfs = lessfs_statfs,
    .release = lessfs_release,
    .fsync = lessfs_fsync,
    .destroy = lessfs_destroy,
    .init = lessfs_init,
};


void usage(char *appName)
{
    char **argv = (char **) malloc(2 * sizeof(char *));
    argv[0] = appName;
    argv[1] = (char *) malloc(3 * sizeof(char));
    memcpy(argv[1], "-h\0", 3);
    fuse_main(2, argv, &lessfs_oper, NULL);
    FUNC;
    printf("\n"
           "-------------------------------------------------------\n"
           "lessfs %s\n"
           "\n"
           "Usage: %s [/path_to_config.cfg] [mount_point] <FUSE OPTIONS>\n"
           "\n"
           "Example :\nmklessfs /etc/lessfs.cfg \nlessfs /etc/lessfs.cfg    /mnt\n"
           "Mounts          /data on /mnt\n\n"
           "High performance example (requires kernel 2.6.26-rc7 or higher and a recent libfuse)\n"
           "lessfs /etc/lessfs.cfg /fuse -o use_ino,readdir_ino,default_permissions,allow_other,big_writes,max_read=131072,max_write=131072\n"
           "-------------------------------------------------------\n",
           VERSION, appName);
    exit(EXIT_USAGE);
}

int verify_kernel_version()
{
   struct utsname un;
   char *begin;
   char *end;
   int count;

   uname(&un);
   begin=un.release;

   for (count = 0; count < 3; count ++){
      end=strchr(begin,'.');
      if ( NULL != end ) {
         end[0]=0;
         end++;
      }
      if ( count == 0 && atoi(begin) < 2 ) return(-1);
      if ( count == 0 && atoi(begin) > 2 ) break;
      if ( count == 1 && atoi(begin) < 6 ) return(-2);
      if ( count == 1 && atoi(begin) > 6 ) break;
      if ( count == 2 && atoi(begin) < 26 ) return(-3);
      begin=end;
   }
   return(0);
}

int main(int argc, char *argv[])
{
    int res;
    char *p, *maxwrite, *maxread;
    char **argv_new = (char **) malloc(argc * sizeof(char *));
    int argc_new = argc - 1;
    FUNC;

    if ((argc > 1) && (strcmp(argv[1], "-h") == 0)) {
        usage(argv[0]);
    }

    if (argc < 3) {
        usage(argv[0]);
    }

    if (-1 == r_env_cfg(argv[1]))
        usage(argv[0]);
    argv_new[0] = argv[0];
    int i;
    for (i = 1; i < argc - 1; i++) {
        argv_new[i] = argv[i + 1];
        if (NULL != strstr(argv[i + 1], "big_writes")) {
            maxwrite = strstr(argv[i + 1], "max_write=");
            maxread = strstr(argv[i + 1], "max_read=");
            if (NULL != maxwrite && NULL != maxread) {
                p = strchr(maxwrite, '=');
                p++;
                BLKSIZE = atoi(p);
                p = strchr(maxread, '=');
                p++;
                if (atoi(p) != BLKSIZE) {
                    LFATAL
                        ("lessfs : Supplied values for max_read and max_write must match.");
                    fprintf(stderr,
                            "Supplied values for max_read and max_write must match.\n");
                    exit(EXIT_SYSTEM);
                }
                if ( BLKSIZE > 4096 && 0 != verify_kernel_version() ) {
                   LFATAL("The kernel used is to old for larger then 4k blocksizes, kernel >= 2.6.26 is required.");
                   exit(EXIT_SYSTEM); 
                }
            } else {
                LFATAL
                    ("lessfs : big_writes specified without max_write or max_read.");
                fprintf(stderr,
                        "big_writes specified without max_write or max_read.\n");
                exit(EXIT_SYSTEM);
            }
        }
    }
    LDEBUG("lessfs : blocksize is set to %u", BLKSIZE);
    signal(SIGHUP, normalExit);
    signal(SIGTERM, normalExit);
    signal(SIGALRM, normalExit);
    signal(SIGINT, normalExit);
    signal(SIGUSR1, libSafeExit);
    signal(SIGSEGV, segvExit);
    if (NULL != getenv("DEBUG"))
        debug = atoi(getenv("DEBUG"));
    res = fuse_main(argc_new, argv_new, &lessfs_oper, NULL);
    free(argv_new);
    return (res);
}
